# coding: utf8
# Copyright (c) Météo France (2017-)
# This software is governed by the CeCILL-C license under French law.
# http://www.cecill.info
""" This module is designed to compute some Fairmode metrics."""

from __future__ import print_function
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import evaltools as evt
from evaltools.evaluator import Evaluator


class FairmodeEvaluator(Evaluator):
    """ Class designed to compute some Fairmode metrics.

    This class inherits from evaltools.evaluator.Evaluator.

    Parameters
    ----------
    observations : evaluator.Observations object
        Observations used to construct the Evaluator object.
    simulations : evaluator.Simulations object
        Simulations used to construct the Evaluator object.
    threshold : scalar
        Limit concentration value fixed by air quality policies.
    U : scalar
        $U^{95}_{95,r}$ as defined by FAIRMODE for meaurement
        uncertainty calculation
    alpha : scalar
        $\alpha$ as defined by FAIRMODE for meaurement
        uncertainty calculation
    RV : scalar
        Reference value as defined by FAIRMODE for meaurement
        uncertainty calculation
    perc : scalar
        Selected percentile value used in the calculation of FAIRMODE's
        modeling perfomance criteria for high percentiles.
    Np, Nnp : scalar
        Coefficients used to compute in FAIRMODE's observation uncertainty
        for annual averages.
    beta : scalar
        $\beta$ as defined by FAIRMODE.
    color : str
        Default color that will be used in plotting functions.

    """
    def __init__(self, observations, simulations, U, alpha, RV,
                 perc, Np, Nnp, beta, threshold=.75, color='k'):
        Evaluator.__init__(self, observations, simulations, color)
        self.threshold = threshold
        self.U = U
        self.alpha = alpha
        self.RV = RV
        self.perc = perc
        self.Np = Np
        self.Nnp = Nnp
        self.beta = beta

    def RMSU(self, threshold=0.75, forecastDay=0):
        """ Root mean square of measurement uncertainty.

        Parameters
        ----------
        threshold : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        forecastDay : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
        pandas.Series
            Series with index corresponding to the object stations
            and containing root mean square of measurement uncertainty
            for each station.

        """
        obs = self.get_obs(forecastDay=forecastDay)
        mu = self.U*np.sqrt((1-self.alpha**2)*obs**2+(self.alpha*self.RV)**2)
        rmsu = evt.scores.stats2D(
            mu, pd.DataFrame(0., index=obs.index, columns=obs.columns),
            score_list=['RMSE'], axis=0, threshold=threshold, keep_nan=True)
        rmsu.columns = ['RMSU']
        return rmsu.RMSU

    def MQI(self, threshold=0.75, forecastDay=0):
        """ Modelling quality incator.

        Parameters
        ----------
        threshold : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        forecastDay : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
        pandas.Series
            Series with index corresponding to the object stations
            and containing modelling quality incator for each station.

        """
        obs = self.get_obs(forecastDay=forecastDay)
        sim = self.get_sim(forecastDay=forecastDay)
        rmse = evt.scores.stats2D(obs, sim, score_list=['RMSE'], axis=0,
                                  threshold=threshold, keep_nan=True)
        rmsu = self.RMSU(threshold=threshold, forecastDay=forecastDay)
        MQI = rmse.RMSE/(self.beta*rmsu)
        return MQI

    def MQI90(self, threshold=0.75, forecastDay=0):
        """ 90th percentile of modelling quality indicator.

        Parameters
        ----------
        threshold : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        forecastDay : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
        float
            90th percentile of modelling quality incator.

        """
        MQI = self.MQI(threshold=threshold, forecastDay=forecastDay)
        MQI = MQI.dropna().sort_values()
        if len(MQI) == 0:
            return np.nan
        stat90 = int(len(MQI)*0.9)
        dist = len(MQI)*0.9 - stat90
        return MQI.iloc[stat90-1] + (MQI.iloc[stat90]-MQI.iloc[stat90-1])*dist

    def Y90(self, availability_ratio=0.75, forecastDay=0):
        """ 90th percentile of MQI for yearly average model results.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        forecastDay : int
            Forecast day corresponding to the data used in the calculation.

        Returns
        -------
        float
            90th percentile of modelling quality incator
            for yearly average model results.

        """
        obs = self.get_obs(forecastDay=forecastDay)
        sim = self.get_sim(forecastDay=forecastDay)
        scores = evt.scores.stats2D(
            obs, sim,
            score_list=['MeanBias', 'obs_mean', 'obs_percentile',
                        'sim_percentile'],
            axis=0, threshold=float(availability_ratio),
            keep_nan=True,
            perc=.9)
        U95 = self.U*np.sqrt((1-self.alpha**2)*(scores['obs_mean']**2) /
                             self.Np + (self.alpha**2)*(self.RV**2)/self.Nnp)
        RMSu_ = np.sqrt(np.nanmean(U95**2))
        MQI = scores['MeanBias'].abs()/(self.beta*RMSu_)
        MQI = MQI.dropna().sort_values()
        if len(MQI) == 0:
            return np.nan
        stat90 = int(len(MQI)*0.9)
        dist = len(MQI)*0.9 - stat90
        return MQI.iloc[stat90-1] + (MQI.iloc[stat90]-MQI.iloc[stat90-1])*dist

    def plot_summary(self, availability_ratio=0.75, forecastDay=0,
                     outputFile=None, file_formats=['png'], title=""):
        """ Summary statistics diagram.

        Assessement summary diagram as described in FAIRMODE guidance
        document on modelling quality objectives and benchmarking.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available on the period required per
            forecast day to compute the scores for each station.
        forecastDay : int
            Forecast day used in the diagram.
        outputFile : str
            File where to save the plot (without extension).
        file_formats : list of str
            List of file extensions.
        title : str
            Diagram title.

        """
        def common_params(ax, xmin, xmax, points, mqi=None, sym=True):
            """ Draw common features to all subplots.

        Parameters
        ----------
        ax : matplotlib axis
            Current subplot.
        xmin, xmax : scalar
            Limit values for the plot
        points : 1D array-like
            Values for the scatter-plot.
        mqi : scalar
            Modeling quality indicator. If < 1 for 90% of the stations,
            mqo is fullfilled (green dot, otherwise red).
        sym : bool
            Must be set to True if the subplot statistical indicator
            can be negative.

            """
            # plot points
            ax.scatter(points, np.ones(len(points)), zorder=10)
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

            # if points live out of limits, plot a point in the dashed area
            if (points > xmax).any():
                ax.scatter([xmax+(xmax-xmin)*0.04], [1], clip_on=False,
                           c=dot_col)
            if (points < xmin).any():
                ax.scatter([xmin-(xmax-xmin)*0.04], [1], clip_on=False,
                           c=dot_col)

            # remove y ticks
            plt.tick_params(axis='y', which='both', left=False)
            ax.set_yticklabels([""])

            # draw a dashed rectangle at the end of the domain
            if sym is True:
                ax.spines['left'].set_visible(False)
                pol = plt.Polygon(
                    xy=[[xmin, 2],
                        [xmin-(xmax-xmin)*0.08, 2],
                        [xmin-(xmax-xmin)*0.08, 0],
                        [xmin, 0]],
                    closed=False, ls='--',
                    clip_on=False,
                    fc='none',
                    edgecolor='k')
                ax.add_patch(pol)
            ax.spines['right'].set_visible(False)
            pol = plt.Polygon(
                xy=[[xmax, 2],
                    [xmax+(xmax-xmin)*0.08, 2],
                    [xmax+(xmax-xmin)*0.08, 0],
                    [xmax, 0]],
                closed=False,
                ls='--',
                clip_on=False,
                fc='none',
                edgecolor='k')
            ax.add_patch(pol)

            # reduce tick size
            ax.tick_params(axis='x', labelsize='small')

            # give more space to y label
            box = ax.get_position()
            ax.set_position([box.x0+box.width*0.15, box.y0, box.width*0.8,
                             box.height*0.4])

            # MQI fulfillment (plot a green or red dot)
            if mqi is not None:
                mqo = np.sum(mqi < 1)/float(len(mqi)) >= 0.9
                col = '#4CFF00'*int(mqo) + 'r'*int(~mqo)
                ax.scatter([xmax+(xmax-xmin)*0.15], [1], clip_on=False,
                           c=col, s=100)

        # scores
        scores = self.stationScores(
            score_list=['MeanBias', 'PearsonR', 'obs_std', 'sim_std',
                        'obs_mean', 'sim_mean'],
            availability_ratio=availability_ratio)['D{}'.format(forecastDay)]
        if scores.isna().all().all():
            print("No valid stations !!!")
            return None
        rmsu = self.RMSU(threshold=float(availability_ratio),
                         forecastDay=forecastDay)

        # plotting
        plt.close('all')
        fig = plt.figure(figsize=(8, 6))
        plt.suptitle(title)
        dot_col = "#1F77B4"
        ymin = 0
        ymax = 2

        # ------- obs mean -------
        ax = fig.add_subplot(811)
        common_params(ax, xmin=0, xmax=100, points=scores['obs_mean'],
                      sym=False)
        ax.set_ylabel("Observed   \nmean ", labelpad=70,
                      rotation='horizontal', verticalalignment='center',
                      size='small')
        ax.text(105, -2, r"$\mu gm^{-3}$")
        ax.text(-37, 3, "-- observations --", rotation='vertical')

        # ------- obs exceedences -------
        ax = fig.add_subplot(812)
        valid_stations = scores.index[~scores[scores.keys()[0]].isna()]
        nb_exceedences = (self.get_obs(forecastDay=forecastDay) >
                          self.threshold)[valid_stations].sum()
        common_params(ax, xmin=0, xmax=100, points=nb_exceedences, sym=False)
        ax.set_ylabel(
            "Observed   \nexceedences \n(> " +
            str(self.threshold) + r" $\mu gm^{-3}$)",
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center',
            size='small')
        ax.text(105, -2, "days")

        # ------- TIME Bias Norm -------
        ax = fig.add_subplot(813)
        x = scores['MeanBias']/(self.beta*rmsu)
        common_params(ax, xmin=-2, xmax=2, points=x, mqi=np.abs(x).dropna())
        ax.set_ylabel(
            "Bias Norm",
            size='small',
            # "$\\frac{Mean\\:bias}{\\beta RMS_U}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        # colored areas
        rect = plt.Rectangle(xy=(-1., 0.), width=2., height=2.,
                             edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7, 0.), width=1.4, height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # ------- TIME Corr Norm -------
        ax = fig.add_subplot(814)
        x = (2.*scores['obs_std']*scores['sim_std']*(
                1. - scores['PearsonR'])) / (self.beta*rmsu)**2
        common_params(ax, xmin=0, xmax=2, points=x, mqi=x.dropna(), sym=False)
        ax.set_ylabel(
            "1-R Norm",
            size='small',
            # "$\\frac{2\\sigma_O\\sigma_M(1-R)}{\\beta^2 RMS_U^2}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        ax.text(-37/50., 7.5, "----------------- time -----------------",
                rotation='vertical')
        # colored areas
        rect = plt.Rectangle(xy=(0., 0.), width=1., height=2.,
                             edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(0., 0.), width=.5, height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # ------- TIME StdDev Norm -------
        ax = fig.add_subplot(815)
        x = (scores['sim_std']-scores['obs_std'])/(self.beta*rmsu)
        common_params(ax, xmin=-2, xmax=2, points=x, mqi=np.abs(x).dropna())
        ax.set_yticks([1.])
        ax.set_ylabel(
            "StDev Norm",
            size='small',
            # "$\\frac{\\sigma_M-\\sigma_O}{\\beta RMS_U}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        # colored areas
        rect = plt.Rectangle(xy=(-1., 0.), width=2., height=2.,
                             edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7, 0.), width=1.4, height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # ------- Hperc -------
        ax = fig.add_subplot(816)
        obs = self.get_obs(forecastDay=forecastDay)
        sim = self.get_sim(forecastDay=forecastDay)
        perc_values = evt.scores.stats2D(
            obs, sim,
            score_list=['obs_percentile', 'sim_percentile'],
            axis=0, threshold=float(availability_ratio),
            keep_nan=True,
            perc=self.perc)
        U95 = self.U*np.sqrt((1-self.alpha**2) *
                             (perc_values['obs_percentile']**2) +
                             (self.alpha**2)*(self.RV**2))
        Hperc = (perc_values['sim_percentile'] -
                 perc_values['obs_percentile']) / (self.beta*U95)
        common_params(ax, xmin=-2, xmax=2, points=Hperc,
                      mqi=np.abs(Hperc).dropna())
        ax.set_ylabel(
            "Hperc Norm",
            size='small',
            # "$\\frac{M_{perc}-O_{perc}}{\\beta U_{95}(O_{perc})}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        # colored area
        rect = plt.Rectangle(xy=(-1., 0.), width=2., height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # ------- SPACE Corr Norm -------
        ax = fig.add_subplot(817)
        sim = scores['sim_mean']
        obs = scores['obs_mean']
        corr = ((np.nanmean((obs-np.nanmean(obs))*(sim-np.nanmean(sim)))) /
                (np.nanstd(obs)*np.nanstd(sim)))
        U95 = self.U*np.sqrt((1-self.alpha**2)*(obs**2)/self.Np +
                             (self.alpha**2)*(self.RV**2)/self.Nnp)
        RMSu_ = np.sqrt(np.nanmean(U95**2))
        x = ((2.*np.nanstd(obs)*np.nanstd(sim)*(1. - corr)) /
             (self.beta*RMSu_)**2)
        common_params(ax, xmin=0, xmax=2, points=np.array([x]), sym=False,
                      mqi=np.array([x]))
        ax.set_ylabel(
            "1-R Norm",
            size='small',
            # "$\\frac{2\\sigma_\\bar{O}\\sigma_\\bar{M}(1-R)}" +
            # "{\\beta^2 RMS_\\bar{U}^2}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        ax.text(-37/50., 2, "------ space ------", rotation='vertical')
        # colored areas
        rect = plt.Rectangle(xy=(0., 0.), width=1., height=2.,
                             edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(0., 0.), width=.5, height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # ------- SPACE StDev Norm -------
        ax = fig.add_subplot(818)
        x = (np.nanstd(sim)-np.nanstd(obs))/(self.beta*RMSu_)
        common_params(ax, xmin=-2, xmax=2, points=np.array([x]),
                      mqi=np.abs(np.array([x])))
        ax.set_ylabel(
            "StDev Norm",
            size='small',
            # "$\\frac{\\sigma_\\bar{M}-\\sigma_\\bar{O}}" +
            # "{\\beta RMS_\\bar{U}}$",
            # size='large',
            labelpad=70,
            rotation='horizontal',
            verticalalignment='center')
        ax.annotate(
            '{valid}/{all} valid stations'.format(
                valid=len(valid_stations), all=len(self.stations)),
            xy=(1, 0), xycoords='axes fraction', fontsize='large',
            xytext=(40, -20), textcoords='offset points', ha='right',
            va='top')
        # colored areas
        rect = plt.Rectangle(xy=(-1., 0.), width=2., height=2.,
                             edgecolor='none', fc='#FFA500')
        ax.add_patch(rect)
        rect = plt.Rectangle(xy=(-.7, 0.), width=1.4, height=2.,
                             edgecolor='none', fc='#4CFF00')
        ax.add_patch(rect)

        # save figure
        if outputFile is None:
            plt.show()
        else:
            for ext in file_formats:
                plt.savefig((outputFile+".{ext}").format(ext=ext))

    def plot_targetDiagram(self, availability_ratio=0.75, forecastDay=0,
                           label=None, color=None, title="", outputFile=None,
                           file_formats=['png'], outputCSV=None,
                           list_stations_ouside_target=True):
        """ Target diagram.

        Assessement target diagram as described in FAIRMODE guidance document
        on modelling quality objectives and benchmarking.

        Parameters
        ----------
        availability_ratio : float
            Minimal rate of data available on the period required per forecast
            day to compute the scores for each station.
        forecastDay : int
            Forecast day used in the diagram.
        label : str
            Label for the legend.
        color : None or str
            Point color.
        title : str
            Diagram title.
        outputFile : str
            File where to save the plot (without extension).
        file_formats : list of str
            List of file extensions.
        outputCSV : str or None
            File where to save the data. The File name must contain {model}
            instead of the model name (so that one file is written for each
            object showed on the graph).
        list_stations_ouside_target : bool
            If True, codes of stations outside the target are written on the
            side of the graph. The option works when only one object is used
            for computation.

        """
        if label is None:
            label = self.model
        if color is None:
            color = self.color
        plot_targetDiagram(
            [self], availability_ratio=availability_ratio,
            forecastDay=forecastDay, labels=[label], colors=[color],
            title=title, outputFile=outputFile, file_formats=file_formats,
            outputCSV=outputCSV,
            list_stations_ouside_target=list_stations_ouside_target)


def plot_targetDiagram(objects, availability_ratio=0.75, forecastDay=0,
                       labels=None, colors=None, title="", outputFile=None,
                       file_formats=['png'], outputCSV=None,
                       list_stations_ouside_target=True):
    """ Target diagram.

    Assessement target diagram as described in FAIRMODE guidance document
    on modelling quality objectives and benchmarking.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        day to compute the scores for each station.
    forecastDay : int
        Forecast day used in the diagram.
    labels : list of str
        List of labels for the legend (length of the list of labels must be
        equal to the length of the list of objects).
    colors : None or list of str
        Line colors corresponding to each objects.
    title : str
        Diagram title.
    outputFile : str
        File where to save the plot (without extension).
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data. The File name must contain {model}
        instead of the model name (so that one file is written for each
        object showed on the graph).
    list_stations_ouside_target : bool
        If True, codes of stations outside the target are written on the
        side of the graph. The option works when only one object is used
        for computation.

    """
    if colors is None:
        colors = [obj.color for obj in objects]
    if labels is None:
        labels = [obj.model for obj in objects]

    plt.close('all')
    fig = plt.figure(figsize=evt.plotting.figsize)
    ax = fig.add_subplot(111)

    # axes
    ax.set_xlim(-2, 2)
    ax.set_ylim(-2, 2)
    major_ticks = np.arange(-2, 3, 1)
    minor_ticks = np.arange(-2, 2.1, 0.1)
    ax.set_xticks(major_ticks)
    ax.set_xticks(minor_ticks, minor=True)
    ax.set_xticklabels([2, 1, 0, 1, 2])
    ax.set_yticks(major_ticks)
    ax.set_yticks(minor_ticks, minor=True)
    ax.yaxis.set_ticks_position('both')
    ax.xaxis.set_ticks_position('both')
    ax.set_aspect('equal')
    plt.xlabel("CRMSE / $\\beta RMS_U$")
    plt.ylabel("Mean Bias / $\\beta RMS_U$")

    # target lines
    for coords in [(-2, 2, -2, 2), (-2, 2, 2, -2)]:
        plt.plot([coords[0], coords[1]], [coords[2], coords[3]], 'k', lw=0.5)
    plt.text(-1.9, 0, "R", color='k', verticalalignment='center',
             horizontalalignment='left', size='medium')
    plt.text(1.9, 0, "SD", color='k', verticalalignment='center',
             horizontalalignment='right', size='medium')
    plt.text(0, 1.9, "Mean bias > 0", color='k', verticalalignment='top',
             horizontalalignment='center', size='medium')
    plt.text(0, -1.9, "Mean bias < 0", color='k', verticalalignment='bottom',
             horizontalalignment='center', size='medium')

    # target circles
    smax = 1
    levels = 2
    xs, ys = np.meshgrid(np.linspace(-smax, smax), np.linspace(-smax, smax))
    rms = np.sqrt(xs**2 + ys**2)
    contours = ax.contour(xs, ys, rms, levels, colors='k',
                          linestyles='dotted', linewidths=0.5)
    circle = plt.Circle((0, 0), 1, color='#ECECEC', zorder=0)
    ax.add_artist(circle)

    # Make place on the side of the figure
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # scatter plot
    for obj, c, lab in zip(objects, colors, labels):
        rmsu = obj.RMSU(threshold=float(availability_ratio),
                        forecastDay=forecastDay)
        obs = obj.get_obs(forecastDay=forecastDay)
        sim = obj.get_sim(forecastDay=forecastDay)
        sc = evt.scores.stats2D(obs, sim,
                                score_list=['MeanBias', 'CRMSE', 'PearsonR'],
                                axis=0,
                                threshold=float(availability_ratio),
                                keep_nan=True)
        if sc.isna().all().all():
            print("No valid station !!!")
            return None
        # avoid warning when calculating std of all nan vector
        with np.warnings.catch_warnings():
            np.warnings.filterwarnings('ignore',
                                       'Degrees of freedom <= 0 for slice')
            idx_both_nan = np.logical_or(sim.isna(), obs.isna())
            obs = obs.where(~idx_both_nan)
            sim = sim.where(~idx_both_nan)
            stdo = np.std(obs, axis=0)
            stds = np.std(sim, axis=0)
        signx = (np.abs(stdo - stds) /
                 ((stdo*stds*2*(1-sc.PearsonR))**0.5) > 1)*2 - 1
        x = (sc.CRMSE/(obj.beta*rmsu))*signx
        y = sc.MeanBias/(obj.beta*rmsu)
        MQI90 = str(round(obj.MQI90(threshold=float(availability_ratio),
                                    forecastDay=forecastDay), 3))
        Y90 = str(round(obj.Y90(availability_ratio=float(availability_ratio),
                                forecastDay=forecastDay), 3))
        ax.scatter(
            x, y, marker='^', facecolors='none', edgecolors=c,
            # label=lab + " (MQI = {})".format(
            #     round(obj.MQI90(threshold=float(availability_ratio),
            #                     forecastDay=forecastDay), 3))
            label=lab+"\n($MQI_{90}$ = "+MQI90+", $Y_{90}$ = "+Y90+")",
            zorder=10)

        if outputCSV is not None:
            radius = np.sqrt(x**2+y**2)
            csv_data = pd.DataFrame({'x': x, 'y': y, 'radius': radius})
            csv_data.sort_values('radius', inplace=True, ascending=False)
            csv_data.to_csv(
                outputCSV.format(model=obj.model),
                sep=' ', na_rep='nan', float_format='%g', header=True,
                index=True)

    # legend
    plt.text(2.1, 1.6, "$\\alpha$ = {}".format(objects[0].alpha), color='k',
             verticalalignment='center', horizontalalignment='left',
             size='medium')
    plt.text(2.1, 1.4, "$\\beta$ = {}".format(objects[0].beta), color='k',
             verticalalignment='center', horizontalalignment='left',
             size='medium')
    plt.text(2.1, 1.2, "RV = {}".format(objects[0].RV), color='k',
             verticalalignment='center', horizontalalignment='left',
             size='medium')
    plt.text(2.1, 1, "$U^{RV}_{95,r}$ = " + "{}".format(objects[0].U),
             color='k', verticalalignment='center',
             horizontalalignment='left', size='medium')
    plt.text(2.1, 0.5, "{} stations".format(np.sum(~x.isna())), color='k',
             verticalalignment='center', horizontalalignment='left',
             size='medium')
    # plt.text(2.1, 0.7, "Model : {0}".format(obj.model))
    # plt.text(2.1, 0.5, "Species : {0}".format(obj.species))
    # plt.text(2.1, 0.5, "Period : {0} - {1}".format(
    #     obj.startDate.strftime('%Y%m%d'), obj.endDate.strftime('%Y%m%d')))
    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))

    if list_stations_ouside_target and len(objects) == 1:
        mqi = np.sqrt(x**2+y**2).sort_values(ascending=False)
        idx = mqi > 1
        out_sta = mqi.loc[idx].index
        sep = [" ", "\n"]

        plt.text(2.1, -.25,
                 "{nb} ".format(nb=len(out_sta)) +
                 "station" + "s"*(len(out_sta) not in [0, 1]) +
                 " with MQI > 1" +
                 (":\n"+''.join([a+sep[i % 2]
                                 for i, a in enumerate(out_sta[:24])])
                  )*(len(out_sta) > 0),
                 color='k', verticalalignment='top',
                 horizontalalignment='left', size='small')

    # titles
    plt.title(title, loc='center')

    if outputFile is None:
        plt.show()
    else:
        for ext in file_formats:
            plt.savefig((outputFile+".{ext}").format(ext=ext))
